@@ -12,7 +12,8 @@ from .download import (
     run_series_list,
     convert_miniml,
     build_composite_dataset,
-    search
+    search,
+    pipeline_find_betas_any_source
     )
 
 
@@ -39,12 +40,21 @@ def build_parser():
         action='store_true',
     )
 
-    subparsers = parser.add_subparsers(dest='command') #, required=True)
-    subparsers.required = True # this is a python3.4-3.7 bug; cannot specify in the call above.
+    parser.add_argument(
+        '-d', '--debug',
+        help='Display VERY detailed messages during processing.',
+        action='store_true',
+    )
+
+    subparsers = parser.add_subparsers(dest='command', required=True)
+    #subparsers.required = True # this is a python3.4-3.7 bug; cannot specify in the call above.
 
     process_parser = subparsers.add_parser('process', help='Finds idat files and calculates raw, beta, m_values for a batch of samples.')
     process_parser.set_defaults(func=cli_process)
 
+    download_parser = subparsers.add_parser('beta_bake', help='All encompasing pipeline that will find GEO datasets in any form, download, and convert into a pickled dataframe of beta-values. Just specify the GEO_ID.')
+    download_parser.set_defaults(func=cli_beta_bakery)
+
     download_parser = subparsers.add_parser('download', help='Downloads the specified series from GEO or ArrayExpress.')
     download_parser.set_defaults(func=cli_download)
 
@@ -63,18 +73,15 @@ def build_parser():
     parsed_args, func_args = parser.parse_known_args(sys.argv[1:])
     if parsed_args.verbose:
         logging.basicConfig(level=logging.INFO)
+    elif parsed_args.debug:
+        logging.basicConfig(level=logging.DEBUG)
     else:
         logging.basicConfig(level=logging.WARNING)
 
-    if parsed_args.command is None:
-        parsed_args.command = 'process'
-
     parsed_args.func(func_args)
     return parser
 
 
-
-
 def cli_process(cmd_args):
     parser = DefaultParser(
         prog='methylprep process',
@@ -188,12 +195,28 @@ def cli_process(cmd_args):
         help='If specified, saves an additional "control_probes.pkl" file that contains Control and SNP-I probe data in the data_dir.'
     )
 
+    parser.add_argument(
+        '--poobah',
+        required=False,
+        action='store_true',
+        default=False,
+        help='By default, any beta-values or m-values output will contain all probes. If True, those probes that fail the p-value signal:noise detection are replaced with NaNs in dataframes in beta_values and m_value output.'
+    )
+
+    parser.add_argument(
+        '--export_poobah',
+        required=False,
+        action='store_true',
+        default=False,
+        help='If specified, exports a pickled dataframe of the poobah p-values per sample.'
+    )
+
     parser.add_argument(
         '-a', '--all',
         required=False,
         action='store_true',
         default=False,
-        help='If specified, saves everything: (beta_values.pkl, m_value.pkl, control_probes.pkl, CSVs for each sample, uncluding uncorrected raw values, and meta data). This overrides individual CLI settings.'
+        help='If specified, saves everything: (beta_values.pkl, m_value.pkl, control_probes.pkl, CSVs for each sample, uncluding uncorrected raw values, and meta data, and poobah_values.pkl). And removes failed probes using sesame pOOBah method from these files. This overrides individual CLI settings.'
     )
 
     args = parser.parse_args(cmd_args)
@@ -209,8 +232,10 @@ def cli_process(cmd_args):
         args.m_value = True
         args.uncorrected = True
         args.save_control = True
-        args.no_export = False
-        args.no_meta_export = False
+        args.no_export = True
+        args.no_meta_export = True
+        args.poobah = True
+        args.export_poobah = True
 
     run_pipeline(
         args.data_dir,
@@ -227,9 +252,85 @@ def cli_process(cmd_args):
         meta_data_frame=args.no_meta_export, # flag flips here
         bit=args.bit,
         save_control=args.save_control,
+        poobah=args.poobah,
+        export_poobah=args.export_poobah,
         )
 
 
+def cli_beta_bakery(cmd_args):
+    parser = DefaultParser(
+        prog='methylprep download',
+        description='Download and process a public dataset, either from GEO or ArrayExpress'
+    )
+
+    parser.add_argument(
+        '-i', '--id',
+        required=True,
+        type=str,
+        help='GEO_ID of the dataset to download',
+    )
+
+    parser.add_argument(
+        '-d', '--data_dir',
+        required=False,
+        type=Path,
+        help='Folder where series data will appear.',
+    )
+
+    parser.add_argument(
+        '-v', '--verbose',
+        required=False,
+        action='store_true',
+        help='if specified, this will turn on more verbose processing messages.',
+    )
+
+    parser.add_argument(
+        '-s', '--save_source',
+        required=False,
+        action='store_true',
+        help='if specified, this will retain .idat and/or -tbl-1.txt files used to generate beta_values dataframe pkl files.',
+    )
+
+    parser.add_argument(
+        '-b', '--bucket',
+        required=False,
+        type=str,
+        help='AWS S3 bucket where downloaded files are stored',
+    )
+
+    parser.add_argument(
+        '-e', '--efs',
+        required=False,
+        type=str,
+        help='AWS elastic file system name, for lambda or AWS batch processing',
+    )
+
+    parser.add_argument(
+        '-p', '--processed_bucket',
+        required=False,
+        type=str,
+        help='AWS S3 bucket where final files are saved',
+    )
+
+    parser.add_argument(
+        '-n', '--no_clean',
+        required=False,
+        default=False,
+        action="store_true",
+        help='If specified, this LEAVES processing and raw data files in temporary folders. By default, these files are removed during processing, and useful files moved to data_dir.',
+    )
+
+    args = parser.parse_args(cmd_args)
+    args.project_name = args.id
+    delattr(args,'id')
+    if args.no_clean == True:
+        args.clean = False
+    else:
+        args.clean = True
+    delattr(args,'no_clean')
+    pipeline_find_betas_any_source(**vars(args))
+
+
 def cli_download(cmd_args):
     parser = DefaultParser(
         prog='methylprep download',
@@ -253,7 +354,7 @@ def cli_download(cmd_args):
         '-l', '--list',
         required=False,
         type=Path,
-        help='List of series IDs (can be either GEO or ArrayExpress)',
+        help='Filename of a text file containing a list of series IDs. IDs can be either GEO or ArrayExpress. One ID on each line',
         )
 
     parser.add_argument(
@@ -272,19 +373,24 @@ def cli_download(cmd_args):
     )
 
     parser.add_argument(
-        '-c', '--no_clean',
+        '-n', '--no_clean',
         required=False,
         action="store_false",
         help='Leave processing and raw data files in folders. By default, these files are removed during processing.'
     )
 
     args = parser.parse_args(cmd_args)
+    if args.no_clean == True:
+        args.clean = False
+    else:
+        args.clean = True
+    delattr(args,'no_clean')
 
     if args.id:
         if args.batch_size:
-            run_series(args.id, args.data_dir, dict_only=args.dict_only, batch_size=args.batch_size, clean=args.no_clean)
+            run_series(args.id, args.data_dir, dict_only=args.dict_only, batch_size=args.batch_size, clean=args.clean)
         else:
-            run_series(args.id, args.data_dir, dict_only=args.dict_only, clean=args.no_clean)
+            run_series(args.id, args.data_dir, dict_only=args.dict_only, clean=args.clean)
     elif args.list:
         if args.batch_size:
             run_series_list(args.list, args.data_dir, dict_only=args.dict_only, batch_size=args.batch_size)
@@ -334,6 +440,15 @@ based on the associated meta data."""
         action="store_true",
         help="[experimental]: If flagged, this will scan the `data_dir` and remove all idat files that are not in the filtered samplesheet, so they won't be processed.",
     )
+
+    parser.add_argument(
+        '-o', '--dont_download',
+        required=False,
+        default=True, # passed in download=True below, unless user overrides with --dont_download
+        action="store_false",
+        help='By default, this will first look at the local filepath (--data-dir) for `GSE..._family.xml` files. IF this is specified, it wont later look online to download the file. Sometimes a series has multiple files and it is easier to download, extract, and point this parser to each file instead.',
+    )
+
     args = parser.parse_args(cmd_args)
     if not args.id:
         raise KeyError("You must supply a GEO id like `GSE123456`.")
@@ -342,7 +457,8 @@ based on the associated meta data."""
         data_dir=args.data_dir,
         extract_controls=args.control,
         require_keyword=args.keyword,
-        sync_idats=args.sync_idats)
+        sync_idats=args.sync_idats,
+        download_it=args.dont_download)
 
 
 def cli_composite(cmd_args):
